"use strict";
exports.__esModule = true;
var common_util_1 = require("../common-util");
var assign = require("object-assign");
var NO_ID = '_noid_';
/**
 * IMPORTANT: Usage of [multi](https://github.com/ashaffer/redux-multi) middleware or a lib of similar nature is required
 */
var FETCH = 'FETCH';
var ACTION = 'ACTION';
var SUCCESS = 'SUCCESS';
var ERROR = 'ERROR';
var REST_METHODS = [{
        name: 'Fetch',
        method: 'GET',
        fetchOrAction: FETCH
    }, {
        name: 'Get',
        method: 'GET',
        fetchOrAction: ACTION,
        replaceModel: true
    }, {
        name: 'Delete',
        method: 'DELETE',
        isDelete: true,
        fetchOrAction: ACTION
    }, {
        name: 'Put',
        method: 'PUT',
        fetchOrAction: ACTION,
        replaceModel: true
    }, {
        name: 'Patch',
        method: 'PATCH',
        fetchOrAction: ACTION,
        replaceModel: true
    }, {
        name: 'Post',
        method: 'POST',
        fetchOrAction: ACTION,
        replaceModel: true
    }];
var STATIC_METHODS = [{
        name: 'LocalPut',
        action: 'SET',
        mapParamTo: 'result'
    }, {
        name: 'ModelData',
        action: 'SET_DATA',
        mapParamTo: 'data'
    }, {
        name: 'LocalDelete',
        action: 'DELETE',
        payload: {}
    }];
function default_1(options) {
    common_util_1.checkRequiredOptions(['actionPrefix', 'entityType'], options);
    var actionPrefix = options.actionPrefix, entityType = options.entityType, normalize = options.normalize, bubbleUp = options.bubbleUp, debug = options.debug;
    var log = common_util_1.logger("action-creator-redux-effects \"" + entityType + "\"");
    // return a callback handler which includes the provided id in the payload as a top level attribute
    function asyncResponseAction(_a) {
        var fetchOrAction = _a.fetchOrAction, type = _a.type, id = _a.id, actionId = _a.actionId, replaceModel = _a.replaceModel, bubbleUp = _a.bubbleUp, isDelete = _a.isDelete, schema = _a.schema, resolver = _a.resolver, formatter = _a.formatter, reduxAction = _a.reduxAction, clearAfter = _a.clearAfter;
        return function (response) {
            // response is assumed to be in [normalize](https://github.com/paularmstrong/normalize) format of
            // {result: _id_, entities: {_entityType_: {_id_: ...}}}
            var payload = response.value;
            var isActionType = actionId && !replaceModel;
            var typeKey = isActionType ? 'response' : 'result';
            var formatterOptions = {
                id: id,
                actionId: actionId,
                entityType: entityType
            };
            if (type === SUCCESS) {
                payload = formatSuccessPayload({
                    payload: payload,
                    formatter: formatter,
                    formatterOptions: formatterOptions,
                    type: typeKey,
                    schema: schema,
                    normalize: normalize
                });
            }
            else {
                payload = { response: response };
            }
            if (!payload) {
                payload = {};
            }
            payload.id = id;
            if (actionId) {
                payload.actionId = actionId;
            }
            if (isDelete) {
                payload["delete"] = true;
            }
            if (typeof bubbleUp !== 'undefined') {
                payload.bubbleUp = bubbleUp;
            }
            var actionType = actionPrefix + "_" + fetchOrAction + "_" + type;
            var action = createAction(actionType, payload);
            var genericAction = createAction(fetchOrAction + "_" + type, payload);
            if (debug) {
                log("triggering " + actionType + " with ", action);
            }
            var rtn = [genericAction, action, function (dispatch) {
                    if (resolver) {
                        resolver(payload);
                    }
                    if (reduxAction) {
                        if (typeof reduxAction === 'function') {
                            reduxAction = reduxAction(payload);
                        }
                        if (reduxAction) {
                            dispatch(reduxAction);
                        }
                    }
                }];
            if (clearAfter) {
                // requires `redux-thunk`
                rtn.push(function (dispatch) {
                    setTimeout(function () {
                        if (debug) {
                            log("action timeout " + entityType + ":" + id);
                        }
                        dispatch(asyncResponseAction({
                            fetchOrAction: fetchOrAction,
                            bubbleUp: bubbleUp,
                            type: 'CLEAR',
                            id: id
                        }));
                    }, clearAfter);
                });
            }
            return rtn;
        };
    }
    var rtn = {};
    STATIC_METHODS.forEach(function (options) {
        var mapParamTo = options.mapParamTo, payload = options.payload, action = options.action, name = options.name;
        rtn["create" + name + "Action"] = function (id, p1, options) {
            options = options || {};
            // if an actionId is provided for a createFetchAction, they really mean createGetAction
            if (options.actionId && name === 'Fetch') {
                return rtn.createGetAction(id, p1, options);
            }
            var formatter = options.formatter, schema = options.schema;
            var _payload = payload;
            if (!_payload) {
                if (formatter || schema) {
                    _payload = formatSuccessPayload({
                        payload: payload,
                        formatter: formatter,
                        formatterOptions: { id: id },
                        type: mapParamTo || 'result',
                        schema: schema,
                        normalize: normalize
                    });
                }
                else if (mapParamTo) {
                    _payload = {};
                    _payload[mapParamTo] = p1;
                }
            }
            _payload.id = id;
            return createAction(actionPrefix + "_" + action, _payload);
        };
    });
    REST_METHODS.forEach(function (options) {
        var fetchOrAction = options.fetchOrAction, method = options.method;
        var isDelete = options["delete"];
        var _replaceModel = options.replaceModel;
        /**
         * return the action to be dispatched when an XHR-based action should be taken on a model/REST document
         * - ACTION_SUCCESS_{entityType}: the data was retrieved successfully
         * - ACTION_ERROR_{entityType}: there was an error with the request
         * - ACTION_PENDING_{entityType}: an XHR request was submitted
         * parameters include
         * - entityType: the entityType key used for all of the event type values
         * - id: the model id (to be added to the payloads for the reducer)
         * - url: the endpoint URI
         * - payload: [effects-fetch payload](https://github.com/redux-effects/redux-effects-fetch#creating-a-user)
         * - clearAfter: clear the action results after N milliseconds (optional)
         */
        rtn["create" + options.name + "Action"] = function (options) {
            var id = options.id, actionId = options.actionId, url = options.url, params = options.params, schema = options.schema, formatter = options.formatter, replaceModel = options.replaceModel, successAction = options.successAction, errorAction = options.errorAction, clearAfter = options.clearAfter;
            var _bubbleUp = typeof options.bubbleUp === 'undefined' ? bubbleUp : options.bubbleUp;
            var _delete = typeof options["delete"] !== 'undefined' ? options["delete"] : isDelete;
            replaceModel = typeof replaceModel === 'undefined' ? _replaceModel : replaceModel;
            if (typeof actionId === 'undefined' && fetchOrAction === ACTION) {
                actionId = method;
            }
            if (id === false) {
                id = NO_ID;
            }
            params = assign({}, params, {
                method: method
            });
            var resolve, reject, promise;
            if (typeof Promise !== 'undefined') {
                promise = new Promise(function (_resolve, _reject) {
                    resolve = function (payload) {
                        _resolve(payload);
                    };
                    reject = _reject;
                });
            }
            var pendingAction = createPendingAction({
                id: id,
                fetchOrAction: fetchOrAction,
                actionPrefix: actionPrefix,
                actionId: actionId,
                bubbleUp: _bubbleUp
            });
            var fetchAction = fetch(url, params);
            var composedAction = bind(fetchAction, asyncResponseAction({
                fetchOrAction: fetchOrAction,
                type: SUCCESS,
                id: id,
                actionId: actionId,
                replaceModel: replaceModel,
                isDelete: _delete,
                bubbleUp: _bubbleUp,
                schema: schema,
                formatter: formatter,
                resolver: resolve,
                reduxAction: successAction,
                clearAfter: clearAfter
            }), asyncResponseAction({
                fetchOrAction: fetchOrAction,
                type: ERROR,
                id: id,
                actionId: actionId,
                bubbleUp: _bubbleUp,
                formatter: formatter,
                resolver: reject,
                reduxAction: errorAction,
                clearAfter: clearAfter
            }));
            var rtn = [composedAction, pendingAction];
            rtn.promise = promise;
            if (debug) {
                log("creating XHR action (" + id + ":" + actionId + ") with:\n\t", rtn);
            }
            return rtn;
        };
    });
    return rtn;
}
exports["default"] = default_1;
function formatSuccessPayload(_a) {
    var payload = _a.payload, formatter = _a.formatter, formatterOptions = _a.formatterOptions, type = _a.type, schema = _a.schema, normalize = _a.normalize;
    formatter = formatter || defaultFormat(type);
    var _payload = formatter(payload, formatterOptions);
    if (!_payload || _payload === payload) {
        payload = defaultFormat(type)(payload);
    }
    else {
        payload = _payload;
    }
    if (schema && normalize && payload) {
        payload = assign(payload, normalize(payload.result, schema));
    }
    return payload;
}
// create a dispatchable action that represents a pending model/REST document action
function createPendingAction(_a) {
    var actionPrefix = _a.actionPrefix, id = _a.id, actionId = _a.actionId, fetchOrAction = _a.fetchOrAction, bubbleUp = _a.bubbleUp;
    var type = fetchOrAction;
    var payload = { id: id };
    if (typeof bubbleUp !== 'undefined') {
        payload.bubbleUp = bubbleUp;
    }
    if (actionId) {
        payload.actionId = actionId;
    }
    return createAction(actionPrefix + "_" + type + "_PENDING", payload);
}
// return an action using the given type and payload
function createAction(type, response) {
    return {
        type: type,
        payload: response
    };
}
// duplicate a little redix-effects/redux-effects-fetch/redux-actions code so this lib is not dependant on either lib
var EFFECT_COMPOSE = 'EFFECT_COMPOSE';
var EFFECT_FETCH = 'EFFECT_FETCH';
function bind(action) {
    var args = [];
    for (var _i = 1; _i < arguments.length; _i++) {
        args[_i - 1] = arguments[_i];
    }
    return {
        type: EFFECT_COMPOSE,
        payload: action,
        meta: {
            steps: [args]
        }
    };
}
function fetch(url, params) {
    if (url === void 0) { url = ''; }
    if (params === void 0) { params = {}; }
    return {
        type: EFFECT_FETCH,
        payload: {
            url: url,
            params: params
        }
    };
}
function defaultFormat(type) {
    return function (value) {
        var rtn = {};
        rtn[type] = value;
        return rtn;
    };
}
