"use strict";
exports.__esModule = true;
var React = require("react");
var common_util_1 = require("./common-util");
var createReactClass = require("create-react-class");
var model_1 = require("./model");
var assign = require("object-assign");
var NO_ID = '_noid_';
/**
 * smart component utility function to ensure a component-specific model will be fetched if it doesn't
 * exist in the store.  A `fetch` prop value is expected to be provided with `mapDispatchToProps`
 * - Component: the "dumb" component
 */
function modelProvider(options) {
    var debug = options.debug, model = options.model, models = options.models, _a = options.entitiesProp, entitiesProp = _a === void 0 ? 'entities' : _a;
    if (!models && !model) {
        throw new Error('either `models` or `model` are required');
    }
    var log = common_util_1.logger('model-provider');
    // organize up our model and collection requirements
    var _models = [];
    if (model) {
        _models.push(organizeProps(model));
    }
    else {
        for (var i = 0; i < models.length; i++) {
            var _model = models[i];
            _models.push(organizeProps(_model));
        }
    }
    // optimize for deep fetching
    _models.forEach(function (source) {
        var data = source.fetchOptions;
        if (typeof data === 'object') {
            for (var key in data) {
                if (data.hasOwnProperty(key)) {
                    if (typeof data[key] === 'string') {
                        data[key] = data[key].split('.');
                    }
                }
            }
        }
    });
    function maybeFetchModels(props, prevProps) {
        var self = this;
        var state = self.state;
        _models.forEach(function (options, index) {
            // see if the model has been set already
            var model = getPropValue(options.propName, props);
            var fetchFunction = getPropValue(options.fetchProp, props);
            if (fetchFunction && !model) {
                var id = getModelId(props, options);
                if (id) {
                    var prevId = state.fetched[index];
                    var isDifferentId = prevId !== id;
                    var modelCache = self.state.modelCache;
                    var isForceFetchFunction = typeof options.forceFetch === 'function';
                    if (isDifferentId || isForceFetchFunction) {
                        // we may need to fetch
                        var modelOptions = assign({}, options, {
                            id: id,
                            entities: props[entitiesProp]
                        });
                        var model_2 = model_1["default"].fromCache(modelOptions, modelCache);
                        var isFetchPending = model_2 && model_2.isFetchPending();
                        var isFetchError = model_2 && model_2.fetchError();
                        var value = model_2 && model_2.value();
                        var shouldFetch = model_2 && !isFetchPending && !isFetchError && !value;
                        if (!shouldFetch && !isFetchPending && (!shouldFetch || isForceFetchFunction)) {
                            // see if we should force it
                            shouldFetch = isForceFetchFunction
                                ? options.forceFetch(id, model_2, props, prevProps) : isDifferentId && options.forceFetch;
                        }
                        if (shouldFetch) {
                            if (debug) {
                                log("fetching model with id \"" + id + "\" data using \"" + options.fetchProp + "\"");
                            }
                            model_1["default"].clearCache(prevId, options.entityType, modelCache);
                            if (id !== undefined && options.entityType) {
                                // seed our state so a model will show up in the props
                                var meta = state.modelCache._meta = state.modelCache._meta || {};
                                var entities = meta[options.entityType] = meta[options.entityType] || {};
                                entities[id] = assign({}, entities[id], {
                                    fetch: {
                                        pending: true
                                    }
                                });
                            }
                            fetchModel(id, props, options);
                            state.fetched[index] = id;
                        }
                        else if (debug) {
                            if (isDifferentId) {
                                log("not fetching model \"" + options.propName + "\"; id changed from " + prevId + " to " + id + " but data exists in state");
                            }
                        }
                    }
                    else if (debug) {
                        log("not fetching model \"" + options.propName + "\" because id value has not changed from " + id);
                    }
                }
                else if (debug) {
                    log("not fetching model \"" + options.propName + "\" because the id \"" + id + "\" was not found in props", props);
                }
            }
        });
    }
    function fetchModel(id, props, options) {
        var fetchOptionsDef = options.fetchOptions;
        var fetchOptions = {};
        if (typeof fetchOptionsDef === 'function') {
            fetchOptions = fetchOptionsDef(props, id);
        }
        else if (fetchOptionsDef) {
            for (var key in fetchOptionsDef) {
                if (fetchOptionsDef.hasOwnProperty(key)) {
                    fetchOptions[key] = getPropValue(fetchOptionsDef[key], props);
                }
            }
        }
        if (debug) {
            log("triggering fetch for model \"" + id + "\" using \"" + options.fetchProp + "\" prop\"");
        }
        var fetchFunc = props[options.fetchProp];
        if (typeof fetchFunc !== 'function') {
            throw new Error("props." + options.fetchProp + " is " + typeof fetchFunc + " but should be a function");
        }
        props[options.fetchProp](id === NO_ID ? false : id, fetchOptions);
    }
    return function (_Component) {
        if (!_Component) {
            throw new Error('Undefined modelProvider component');
        }
        return createReactClass({
            getInitialState: function () {
                return {
                    modelCache: {},
                    fetched: {}
                };
            },
            componentWillMount: function () {
                maybeFetchModels.call(this, this.props);
                triggerIdChanges(undefined, this.props, this.state);
            },
            componentWillReceiveProps: function (props) {
                maybeFetchModels.call(this, props, this.props);
                triggerIdChanges(this.props, props, this.state);
            },
            render: function () {
                var props = this.props;
                return React.createElement(_Component, generateProps(props, this.state), props.children);
            }
        });
        function checkForMissingOrChangedIds(oldProps, newProps) {
            return _models.map(function (options) {
                if (!oldProps) {
                    return {
                        options: options,
                        oldId: undefined,
                        newId: getModelId(newProps, options)
                    };
                }
                var oldId = getModelId(oldProps, options);
                var newId = getModelId(newProps, options);
                if (oldId !== newId) {
                    return {
                        options: options,
                        oldId: undefined,
                        newId: getModelId(newProps, options)
                    };
                }
            }).filter(function (o) { return o; });
        }
        function triggerIdChanges(oldProps, newProps, state) {
            var changelist = checkForMissingOrChangedIds(oldProps, newProps);
            if (changelist.length > 0) {
                var props_1 = generateProps(newProps, state);
                changelist.forEach(function (data) {
                    if (data.options.onIdChange) {
                        var options_1 = data.options;
                        var oldId = data.oldId;
                        var newId = data.newId;
                        data.options.onIdChange(newId, oldId, props_1);
                    }
                });
            }
        }
        function generateProps(origProps, state) {
            var props = assign({}, origProps);
            var modelCache = state.modelCache;
            _models.forEach(function (options) {
                var id = getModelId(props, options);
                if (id) {
                    var modelOptions = assign({}, options, {
                        id: id,
                        entities: props[entitiesProp]
                    });
                    // reuse the same model object if we can
                    var model_3 = model_1["default"].fromCache(modelOptions, modelCache);
                    if (options.idPropName) {
                        setPropValue(props, options.idPropName, id);
                    }
                    setPropValue(props, options.propName, model_3);
                }
            });
            return props;
        }
    };
}
exports["default"] = modelProvider;
function getModelId(props, options) {
    var id = options.id;
    if (id === false || id === NO_ID) {
        return NO_ID;
    }
    else {
        return getPropValue(id, props);
    }
}
function getPropValue(keyParts, parent) {
    if (typeof keyParts === 'undefined') {
        return undefined;
    }
    else if (typeof keyParts === 'function') {
        return keyParts(parent);
    }
    else if (Array.isArray(keyParts)) {
        return common_util_1.deepPropValue(keyParts, parent);
    }
}
function setPropValue(parent, keyParts, value) {
    for (var i = 0; i < keyParts.length; i++) {
        var key = keyParts[i];
        if (i === keyParts.length - 1) {
            parent[key] = value;
        }
        else {
            if (!parent[key]) {
                parent[key] = {};
            }
            parent = parent[key];
        }
    }
}
function organizeProps(options) {
    common_util_1.checkRequiredOptions(['id', 'entityType'], options);
    var id = options.id;
    var rtn = assign({}, options, {
        id: id === false ? NO_ID : typeof id === 'string' ? id.split('.') : id,
        entityType: options.entityType,
        propName: (options.propName || 'model').split('.'),
        idPropName: options.idPropName ? options.idPropName.split('.') : undefined,
        fetchProp: options.fetchProp ? options.fetchProp.split('.') : undefined,
        modelClass: options.modelClass || model_1["default"],
        fetchOptions: typeof options.fetchOptions === 'object' ? assign({}, options.fetchOptions) : options.fetchOptions
    });
    if (!rtn.id) {
        console.warn('id prop *must* be provided to model provider');
    }
    return rtn;
}
