"use strict";
exports.__esModule = true;
var NO_ID = '_noid_';
var Model = /** @class */ (function () {
    function Model(options) {
        this.formatted = false;
        var entities = options.entities;
        var id = this.id = determineId(options.id);
        var entityType = options.entityType;
        if (entities) {
            // allow for root state to be passed
            entities = entities.entities || entities;
        }
        var value = options.value || deepValue(entities, [entityType, id]);
        var meta = options.meta || deepValue(entities, ['_meta', entityType, id]);
        this.entities = entities;
        this._value = value;
        this.options = options;
        this._meta = meta || {};
        this._data = this._meta.data || {};
    }
    Model.prototype.meta = function () {
        return this._meta;
    };
    /**
     * Return the (optionally formatted) model data
     */
    Model.prototype.value = function () {
        var _this = this;
        if (!this.formatted) {
            this.formatted = true;
            var options = this.options;
            if (this._value && options.schema && options.denormalize) {
                this.formattedValue = options.denormalize(this._value, options.schema, this.entities);
            }
            else {
                var formatter = options.formatter;
                this.formattedValue = formatter
                    ? formatter(this._value, options)
                    : this._value;
            }
            var arrayEntrySchema_1 = options.arrayEntrySchema;
            var ArrayEntryModel_1 = Model;
            if (arrayEntrySchema_1 && this.formattedValue) {
                if (arrayEntrySchema_1.model) {
                    ArrayEntryModel_1 = arrayEntrySchema_1.model || Model;
                    arrayEntrySchema_1 = arrayEntrySchema_1.schema;
                }
                this.formattedValue = this.formattedValue.map(function (data) {
                    return new ArrayEntryModel_1({
                        entities: _this.entities,
                        id: arrayEntrySchema_1.getId(data),
                        entityType: arrayEntrySchema_1.key
                    });
                });
            }
        }
        return this.formattedValue;
    };
    /**
     * Return metadata associated with this model
     */
    Model.prototype.data = function () {
        return this._data;
    };
    /**
     * Return truthy if the model has been fetched (`fetched` if fetched and `set` if used with constructor to set value)
     */
    Model.prototype.wasFetched = function () {
        var meta = getMeta(this);
        var rtn = !!meta.fetch && meta.fetch.success;
        if (!rtn && this.value()) {
            rtn = true;
        }
        return !!(rtn || this.value());
    };
    /**
     * Return true if there is not a fetch pending or the model has been sucessfully fetched
     */
    Model.prototype.canBeFetched = function () {
        var meta = getMeta(this);
        var fetchData = meta.fetch;
        var hasValue = typeof this._value === 'function' && this.value();
        if (fetchData) {
            if (fetchData.pending) {
                return false;
            }
            else {
                return !(fetchData.success || !!hasValue);
            }
        }
        else {
            return !hasValue;
        }
    };
    /**
     * Return a truthy (timestamp of when the fetch was initiated) if a fetch is pending
     */
    Model.prototype.isFetchPending = function () {
        var meta = getMeta(this);
        var fetchData = meta.fetch;
        return (fetchData && fetchData.pending) || false;
    };
    /**
     * Return a fetch success result or false
     */
    Model.prototype.fetchSuccess = function () {
        var meta = getMeta(this);
        var fetchData = meta.fetch;
        return (fetchData && fetchData.success) || false;
    };
    /**
     * Return a fetch error result or false
     */
    Model.prototype.fetchError = function () {
        var meta = getMeta(this);
        var fetchData = meta.fetch;
        return (fetchData && fetchData.error) || false;
    };
    /**
     * Return a truthy (timestamp of action initiation) if the action is pending
     * @param {string} id: optinal identifier to see if a specific action is currently in progress
     * @paramm {string} actionId: action id
     */
    Model.prototype.isActionPending = function (actionId) {
        verifyActionId(actionId);
        var meta = getMeta(this);
        var actionData = meta.actions && meta.actions[actionId];
        return (actionData && actionData.pending) || false;
    };
    /**
     * If an action was performed and successful, return { success, error, pending }.  `success` and `error` will be mutually exclusive and will
     * represent the XHR response payload
     * @paramm {string} actionId: action id to only return true if a specific action was performed
     */
    Model.prototype.wasActionPerformed = function (actionId) {
        verifyActionId(actionId);
        var meta = getMeta(this);
        var actionData = meta.actions;
        return actionData && actionData[actionId];
    };
    /**
     * If an action was performed and is an in error state, return the error response
     * @paramm {string} actionId: action id to only return true if a specific action was performed
     * @returns the error response payload
     */
    Model.prototype.actionError = function (actionId) {
        verifyActionId(actionId);
        var meta = getMeta(this);
        var actionData = meta.actions;
        return (actionData && actionData[actionId] && actionData[actionId].error) || null;
    };
    /**
     * If an action was performed and is in success state, return the success response
     * @paramm {string} actionId: action id to only return true if a specific action was performed
     * @returns the success response payload or true if the response was a success
     */
    Model.prototype.actionSuccess = function (actionId) {
        verifyActionId(actionId);
        var meta = getMeta(this);
        var actionData = meta.actions;
        return (actionData && actionData[actionId] && actionData[actionId].success) || null;
    };
    /**
     * Return the number of milis since the last fetch completion (success or error)
     */
    Model.prototype.timeSinceFetch = function (currentTime) {
        var meta = getMeta(this);
        var fetchTime = (meta.fetch && meta.fetch.completedAt);
        return fetchTime ? (currentTime || new Date().getTime()) - fetchTime : -1;
    };
    /**
     * Return a model from the cache object and create one if one does not exist
     */
    Model.fromCache = function (options, cache) {
        if (!cache) {
            return null;
        }
        var id = determineId(options.id);
        var entityType = options.entityType;
        var ModelClass = options.modelClass || Model;
        var entities = options.entities || {};
        // allow for root state to be provided
        entities = entities.entities || entities;
        var cachedEntities = cache[entityType] = cache[entityType] || {};
        var cachedMeta = cache._meta = cache._meta || {};
        var cachedModels = cachedEntities.__models = cachedEntities.__models || {};
        var cachedModel = cachedModels[id];
        var cachedData = getMetaAndValue(id, cache, entityType);
        var checkData = getMetaAndValue(id, entities, entityType);
        if (!cachedModel || cachedData.meta !== checkData.meta || cachedData.value !== checkData.value) {
            // we need to cache and return a new model
            cachedEntities[id] = checkData.value;
            var cachedMetaEntity = cachedMeta[entityType] = cachedMeta[entityType] || {};
            cachedMetaEntity[id] = checkData.meta || cachedMetaEntity[id];
            cachedModel = new ModelClass({
                id: id,
                entityType: options.entityType,
                entities: options.entities,
                meta: cachedMetaEntity[id]
            }); // joe
            cachedModels[id] = cachedModel;
        }
        return cachedModel;
    };
    /**
     * Clear the model referred to by the entity type and id from the cache
     */
    Model.clearCache = function (id, entityType, cache) {
        if (!cache) {
            return;
        }
        id = determineId(id);
        var metaTypes = deepValue(cache, ['_meta', entityType]);
        if (metaTypes) {
            delete metaTypes[id];
        }
        var entityTypes = cache[entityType];
        if (entityTypes) {
            delete entityTypes[id];
        }
        var models = deepValue(cache, [entityType, '__models']);
        if (models) {
            delete models[id];
        }
    };
    return Model;
}());
exports["default"] = Model;
// allow the following to have static accessors
[
    'data', 'wasFetched', 'canBeFetched', 'isFetchPending', 'fetchSuccess', 'fetchError', 'isActionPending',
    'wasActionPerformed', 'actionError', 'actionSuccess', 'timeSinceFetch'
].forEach(function (key) {
    var func = Model.prototype[key];
    Model[key] = function () {
        var meta = arguments[0];
        var args = Array.prototype.slice.call(arguments, 1);
        return func.apply({ __static: { meta: meta } }, args);
    };
});
function determineId(id) {
    return id === false ? NO_ID : id;
}
function getMetaAndValue(id, entities, entityType) {
    return {
        meta: deepValue(entities, ['_meta', entityType, id]) || null,
        value: deepValue(entities, [entityType, id]) || null
    };
}
function deepValue(parent, parts) {
    for (var i = 0; i < parts.length && parent; i++) {
        parent = parent[parts[i]];
    }
    return parent;
}
function verifyActionId(actionId) {
    if (!actionId) {
        throw new Error('action id must be provided');
    }
}
function getMeta(context) {
    return context._meta ? context._meta : context.__static ? context.__static.meta : undefined;
}
